/*

    Copyright (C) 1997 Frithjof Brestrich
                       brestrich@kde.org


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/ 

#include <stdlib.h>
#include <stdio.h>

#include "xf86configfile.h"

XF86ConfigFile::XF86ConfigFile()
{
  ErrorText = "";
  Data      = new tXF86CfgDirectory(NULL);
  File      = NULL;
  Stream    = NULL;
}

XF86ConfigFile::~XF86ConfigFile()
{
  CloseFile();
  delete Data;
}

void XF86ConfigFile::SetErrorText(const QString &txt)
{
  ErrorText = txt;
}

const QString &XF86ConfigFile::LastErrorText()
{
  return ErrorText;
}

bool XF86ConfigFile::OpenFile(const QString &fname, int flags)
{
  File = new QFile(fname);
  if ( !File ) { SetErrorText("cannot create file object"); return FALSE; }

  if (File->open( flags )) { 
    File->at(0);
    Stream = new QTextStream(File);
  } else {
     SetErrorText("cannot connect file object to stream object");
    return FALSE;
  }

  return TRUE;
}

bool XF86ConfigFile::CloseFile()
{
  if (!File || !Stream ) return FALSE;

  File->close();

  delete Stream;
  delete File;
  Stream = NULL;
  File = NULL;

  return TRUE;
}

bool XF86ConfigFile::Load(const QString &fname)
{
  bool b = TRUE;

  if (!OpenFile(fname, IO_ReadOnly )) {
    SetErrorText("Cannot open file for reading.\nCheck permissons and existance!");
    return FALSE;
  }

  while ( File->isOpen() && !Stream->eof() ) {

    if (!ReadSection(Data)) {
      b = FALSE;
      break;
    }
  }

  CloseFile(); // ignore error code

  return b;
}

bool XF86ConfigFile::Store(const QString &fname)
{
  bool b = TRUE;

  if (!OpenFile(fname, IO_Truncate | IO_WriteOnly )) {
    SetErrorText("Cannot open file for writing.\nCheck permissons!");
    return FALSE;
  }

  b = WriteSection(Data);
  
  CloseFile(); // ignore error code

  return b;
}

bool XF86ConfigFile::ReadSection(tXF86CfgDirectory *Dir)
{
  QString s,key,val;
  tXF86CfgEntry *e;
  unsigned sep,space,tab;
  bool   b = TRUE;

  if (!Stream) return FALSE;

  while ( File->isOpen() && !Stream->eof() ) {

    if ( (s = Stream->readLine() ) == NULL ) { 
      SetErrorText("error reading a line");
      b = FALSE; 
      break; 
    }

    s = s.stripWhiteSpace();
    if ( s[0] == '#'  || s.length() == 0 ) {
    
      e = Dir->Data.getLast();

      // remove blank lines which occur more than once
      if ( e==NULL || e->Key!=NULL || e->Value!=NULL || s.length() ) {

        if ( (e = new tXF86CfgEntry) == NULL )  { 
	  SetErrorText("out of memory");
	  b = FALSE; 
	  break; 
	}

        e->IsString = FALSE;
        e->Key = NULL;
        if (s.length()!=0) // empty strings become NULL
	  e->Value.setStr(s);
	else
	  e->Value = NULL;

        Dir->Data.append(e);

      }
    
    } else {
    
      space = s.find(' ',0,FALSE);
      tab   = s.find('\t',0,FALSE);
      sep   = space < tab ? space : tab;
      sep   = sep < s.length() ? sep : s.length();

      key.setStr(s).truncate(sep);
      val = s.mid(sep,s.size()).stripWhiteSpace();

      if ( !stricmp(key,"SubSection") || !stricmp(key,"Section") ) {
        tXF86CfgDirectory *d = NULL;

	if ( (val.length()<2) || (val[0]!='"') || (val[val.length()-1]!='"') ) {
	   SetErrorText("invalid section statement");
	   b = FALSE; 
	   break; 
	}

        if ( (d = new tXF86CfgDirectory(val.mid(1,val.length()-2))) == NULL )  { 
	   SetErrorText("out of memory");
	   b = FALSE; 
	   break; 
	}

        if (!ReadSection(d))  { 
	  // error set by ReadSection()
	  b = FALSE; 
	  break; 
	}

	Dir->Dir.append(d);

      } else if ( !stricmp(key,"EndSection") || !stricmp(key,"EndSubSection") ) {
      
        break;
      
      } else {
      
        if ( (e = new tXF86CfgEntry) == NULL )  { 
	  SetErrorText("out of memory");
	  b = FALSE; 
	  break; 
	}
        
	e->Key.setStr(key);
	if ( val.contains('"') == 2 && val[0] == '"' && val[val.length()-1] == '"' ) {
          e->IsString = TRUE;  
          e->Value = val.mid(1,val.length()-2);
	} else {
          e->IsString = FALSE;
          e->Value.setStr(val); 
	}

        Dir->Data.append(e);
      
      }
    }
  }

  return b;
}

bool XF86ConfigFile::WriteSection(tXF86CfgDirectory *Dir, int Deep)
{
  tXF86CfgDirectory *d;
  tXF86CfgEntry     *e;
  QString           s,buf;
  int               i;

  for (i=1;i<Deep;i++) s+="\t";

  if ( Deep == 1 ) *Stream << "\nSection \"" << Dir->Name << "\"\n";
  else if ( Deep > 1) *Stream << s << "SubSection \"" << Dir->Name << "\"\n";

  s+="\t";
  
  for ( e = Dir->Data.first(); e != NULL ; e = Dir->Data.next() ) {
    if ( e->Key == NULL ) {
      if ( e->Value != NULL ) *Stream << e->Value;
      *Stream << '\n';
    } else {
      if (e->IsString) {
        buf.sprintf("%s%-15s \"%s\"\n",(const char*)s,(const char*)e->Key,(const char*)e->Value);
      } else {
        if ( e->Value == NULL )
          buf.sprintf("%s%s\n",(const char*)s,(const char*)e->Key);
	else 

          buf.sprintf("%s%-15s %s\n",(const char*)s,(const char*)e->Key,(const char*)e->Value);
      }
      *Stream << buf;
    }
  }

  for ( d = Dir->Dir.first(); d != NULL ; d = Dir->Dir.next() ) 
    if (!WriteSection(d,Deep+1)) 
      return FALSE;

  if ( Deep == 1 ) *Stream << "EndSection\n";
  else if ( Deep > 1) *Stream << s.left(s.length()-1) << "EndSubSection\n";

  return TRUE;
}

tXF86CfgEntry* XF86ConfigFile::FindEntry(const char* name, tXF86CfgEntryList* root=NULL)
{
  tXF86CfgEntry *e;
  
  if ( root == NULL )
    root = &(Data->Data);

  for ( e = root->last(); e != NULL ; e = root->prev() ) 
    if ( e->Key != NULL  &&  !stricmp(e->Key,name) ) 
      return e;

  return NULL;
}

tXF86CfgDirectory* XF86ConfigFile::FindSection(const char* name,const char *key,const char* val,
                                               tXF86CfgDirectory* root)
{
  tXF86CfgDirectory *d;
  tXF86CfgEntry     *e;

  if ( root == NULL ) 
    root = Data;

  for ( d = root->Dir.last(); d != NULL ; d = root->Dir.prev() ) {

    if ( d->Name == name ) {
      if ( key != NULL  && val != NULL ) {
        e = FindEntry(key,&(d->Data));
	if (e->Value == val )
	  return d;
      } else
        return d;
    }

  }

  return NULL;
}

bool XF86ConfigFile::SetModeLines(char *vendor, char *model, XF86VidModeLine* &modelines,unsigned count)
{
  tXF86CfgDirectory *d;
  tXF86CfgEntry     *e;
  char              *s;
  unsigned           n;

  d = FindSection("Monitor","VendorName",vendor);
  if ( d == NULL ) { 
    SetErrorText("no entry for specified vendor found"); 
    return FALSE; 
  }

  e = FindEntry("ModelName",&(d->Data));
  if ( e == NULL  || e->Value != model ) { 
    SetErrorText("no entry for specified model found"); 
    return FALSE; 
  }

  // delete all
  for ( e = d->Data.first(); e != NULL ; e = d->Data.next() ) 
    if ( (e->Key!=NULL) && !stricmp(e->Key,"Modeline"))  
      if (d->Data.remove())     
        d->Data.prev();

  for ( n=0 ; n<count ; n++ ) {
    if ( (e = new tXF86CfgEntry) == NULL ) {
      SetErrorText("out of memory"); 
      return FALSE; 
    }
    e->Key.setStr("Modeline");
    e->IsString = FALSE;
    s = modelines[n].GetModeline();
    e->Value.setStr(s);
    free(s);
    d->Data.append(e);
  }
      
  return TRUE;
}

